#if defined (UNITTEST)
#include "StdAfx.h"
#else
#include "local.h"
#endif
#include "RuleTreeNodeP.h"
#include "keymap.h"


struct CSkkRuleTreeNodeOutput*
TSkkRuleTreeNodeOutput_pCreateStringPairInstance (
	LPCWSTR							wstrLeft,
	int								nLeftLen,
	LPCWSTR							wstrRight,
	int								nRightLen)
{
	struct CSkkRuleTreeNodeOutput*	pNode ;
	WCHAR*	pwDest ;
	int		nSize ;

	nSize	= sizeof (struct CSkkRuleTreeNodeOutput) + nLeftLen * sizeof (WCHAR) + nRightLen * sizeof (WCHAR) ;
	if (nSize < sizeof (struct CSkkRuleTreeNodeOutput))
		return	NULL ;
	pNode	= (struct CSkkRuleTreeNodeOutput*) MALLOC (nSize) ;
	if (pNode == NULL)
		return	NULL ;
	pwDest			= (WCHAR*) (pNode + 1) ;
	pNode->m_nType	= NTYPE_STRINGPAIR ;
	if (wstrLeft != NULL) {
		pNode->m_uValue.m_stringpair.m_strHira	= pwDest ;
		memcpy (pwDest, wstrLeft, nLeftLen * sizeof (WCHAR)) ;
		pNode->m_uValue.m_stringpair.m_nHira	= nLeftLen ;
		pwDest	+= nLeftLen ;
	} else {
		pNode->m_uValue.m_stringpair.m_strHira	= NULL ;
		pNode->m_uValue.m_stringpair.m_nHira	= 0 ;
	}
	if (wstrRight != NULL) {
		pNode->m_uValue.m_stringpair.m_strKata	= pwDest ;
		memcpy (pwDest, wstrRight, nRightLen * sizeof (WCHAR)) ;
		pNode->m_uValue.m_stringpair.m_nKata	= nRightLen ;
		pwDest	+= nRightLen ;
	} else {
		pNode->m_uValue.m_stringpair.m_strKata	= NULL ;
		pNode->m_uValue.m_stringpair.m_nKata	= 0 ;
	}
	return	pNode ;
}

struct CSkkRuleTreeNodeOutput*
TSkkRuleTreeNodeOutput_pCreateStringInstance (
	LPCWSTR							wstrString,
	int								nStringLen)
{
	struct CSkkRuleTreeNodeOutput*	pNode ;
	WCHAR*	pwDest ;
	int		nSize ;

	nSize	= sizeof (struct CSkkRuleTreeNodeOutput) + nStringLen * sizeof (WCHAR) ;
	if (nSize < sizeof (struct CSkkRuleTreeNodeOutput))
		return	NULL ;
	pNode	= (struct CSkkRuleTreeNodeOutput*) MALLOC (nSize) ;
	if (pNode == NULL)
		return	NULL ;
	pwDest			= (WCHAR*) (pNode + 1) ;
	pNode->m_nType	= NTYPE_STRING ;
	if (wstrString != NULL) {
		pNode->m_uValue.m_string.m_strValue	= pwDest ;
		memcpy (pwDest, wstrString, nStringLen * sizeof (WCHAR)) ;
		pNode->m_uValue.m_string.m_nValue	= nStringLen ;
	} else {
		pNode->m_uValue.m_string.m_strValue	= NULL ;
		pNode->m_uValue.m_string.m_nValue	= 0 ;
	}
	return	pNode ;
}

struct CSkkRuleTreeNodeOutput*
TSkkRuleTreeNodeOutput_pCreateMacroInstance (
	int								nMacro)
{
	struct CSkkRuleTreeNodeOutput*	pNode ;

	pNode	= (struct CSkkRuleTreeNodeOutput*) MALLOC (sizeof (struct CSkkRuleTreeNodeOutput)) ;
	if (pNode == NULL)
		return	NULL ;
	pNode->m_nType				= NTYPE_MACRO ;
	pNode->m_uValue.m_nMacro	= nMacro ;
	return	pNode ;
}

void
TSkkRuleTreeNodeOutput_vDestroy (
	struct CSkkRuleTreeNodeOutput*	pNode)
{
	if (pNode == NULL)
		return ;
	FREE (pNode) ;
	return ;
}

int
TSkkRuleTreeNodeOutput_iGetType		(
	const struct CSkkRuleTreeNodeOutput*	pNode)
{
	return	(pNode != NULL)? pNode->m_nType : -1 ;
}

BOOL
TSkkRuleTreeNodeOutput_bGetMacro	(
	const struct CSkkRuleTreeNodeOutput*	pNode,
	int*									pnFuncNo)
{
	if (pNode == NULL || pNode->m_nType != NTYPE_MACRO)
		return	FALSE ;
	if (pnFuncNo != NULL)
		*pnFuncNo	= pNode->m_uValue.m_nMacro ;
	return	TRUE ;
}

BOOL
TSkkRuleTreeNodeOutput_bGetString	(
	const struct CSkkRuleTreeNodeOutput*	pNode,
	LPCWSTR*								ppwString,
	int*									pnStringLen)
{
	if (pNode == NULL || pNode->m_nType != NTYPE_STRING)
		return	FALSE ;
	if (ppwString != NULL)
		*ppwString	= pNode->m_uValue.m_string.m_strValue ;
	if (pnStringLen != NULL)
		*pnStringLen	= pNode->m_uValue.m_string.m_nValue ;
	return	TRUE ;
}


BOOL
TSkkRuleTreeNodeOutput_bGetLString	(
	const struct CSkkRuleTreeNodeOutput*	pNode,
	LPCWSTR*								ppwString,
	int*									pnStringLen)
{
	if (pNode == NULL || pNode->m_nType != NTYPE_STRINGPAIR)
		return	FALSE ;
	if (ppwString != NULL)
		*ppwString	= pNode->m_uValue.m_stringpair.m_strHira ;
	if (pnStringLen != NULL)
		*pnStringLen	= pNode->m_uValue.m_stringpair.m_nHira ;
	return	TRUE ;
}


BOOL
TSkkRuleTreeNodeOutput_bGetRString	(
	const struct CSkkRuleTreeNodeOutput*	pNode,
	LPCWSTR*								ppwString,
	int*									pnStringLen)
{
	if (pNode == NULL || pNode->m_nType != NTYPE_STRINGPAIR)
		return	FALSE ;
	if (ppwString != NULL)
		*ppwString	= pNode->m_uValue.m_stringpair.m_strKata ;
	if (pnStringLen != NULL)
		*pnStringLen	= pNode->m_uValue.m_stringpair.m_nKata ;
	return	TRUE ;
}

/*===============================================================
 *	struct CSkkRuleTreeNode public functions
 */
LPCWSTR
TSkkRuleTreeNode_pGetPrefix (
	const struct CSkkRuleTreeNode*			pNode,
	int*									pnLength)
{
	if (pNode == NULL)
		return	NULL ;
	if (pnLength != NULL)
		*pnLength	= pNode->_nPrefix ;
	return	pNode->_pPrefix ;
}

LPCWSTR
TSkkRuleTreeNode_pGetNextState	(
	const struct CSkkRuleTreeNode*			pNode,
	int*									pnLength)
{
	if (pNode == NULL)
		return	NULL ;
	if (pnLength != NULL)
		*pnLength	= pNode->_nNextState ;
	return	pNode->_pNextState ;
}

const struct CSkkRuleTreeNodeOutput*
TSkkRuleTreeNode_pGetOutput		(
	const struct CSkkRuleTreeNode*			pNode)
{
	return	(pNode != NULL)? pNode->_pOutput : NULL ;
}

/*===============================================================
 *	static public methods
 */
struct CSkkRuleTreeNode*
TSkkRuleTreeNode_pCreateInstance (
	int								nChar, 
	LPCWSTR							wstrPrefix,
	int								nPrefixLen,
	LPCWSTR							wstrNext,
	int								nNextLen,
	struct CSkkRuleTreeNodeOutput*	pOutput)
{
	struct CSkkRuleTreeNode*	pNode ;
	int					nSize ;
	WCHAR*				pwDest ;

	nSize	= sizeof (struct CSkkRuleTreeNode) + sizeof (WCHAR) * nPrefixLen + sizeof (WCHAR) * nNextLen ;
	pNode	= (struct CSkkRuleTreeNode*) MALLOC (nSize) ;
	if (pNode == NULL)
		return	NULL ;

	pNode->_nCH	= nChar ;
	pwDest		= (WCHAR*) (pNode + 1) ;
	if (nPrefixLen > 0) {
		pNode->_pPrefix	= pwDest ;
		memcpy (pNode->_pPrefix, wstrPrefix, sizeof (WCHAR) * nPrefixLen) ;
		pwDest			+= nPrefixLen ;
	} else {
		pNode->_pPrefix	= NULL ;
		nPrefixLen		= 0 ;
	}
	pNode->_nPrefix	= nPrefixLen ;

	if (nNextLen > 0) {
		pNode->_pNextState	= pwDest ;
		memcpy (pNode->_pNextState, wstrNext, sizeof (WCHAR) * nNextLen) ;
		pwDest		+= nNextLen ;
	} else {
		pNode->_pNextState	= NULL ;
		nNextLen	= 0 ;
	}
	pNode->_nNextState	= nNextLen ;
	pNode->_pOutput		= pOutput ;
	pNode->_pChild		= NULL ;
	pNode->_pBrother	= NULL ;
	return	pNode ;
}

void
TSkkRuleTreeNode_vDestroy (
	struct CSkkRuleTreeNode*	pNode)
{
	if (pNode == NULL)
		return ;

	TSkkRuleTreeNodeOutput_vDestroy (pNode->_pOutput) ;
	pNode->_pOutput	= NULL ;
	FREE (pNode) ;
	return ;
}

void
TSkkRuleTreeNode_vSetBrother (
	struct CSkkRuleTreeNode*		pRoot,
	struct CSkkRuleTreeNode*		pBrother)
{
	pRoot->_pBrother	= pBrother ;
	return ;
}

void
TSkkRuleTreeNode_vSetChild (
	struct CSkkRuleTreeNode*		pRoot,
	struct CSkkRuleTreeNode*		pChild)
{
	pRoot->_pChild	= pChild ;
	return ;
}

void
TSkkRuleTreeNode_vSetOutput (
	struct CSkkRuleTreeNode*		pNode,
	struct CSkkRuleTreeNodeOutput*	pOutput)
{
	if (pNode == NULL)
		return ;
	if (pNode->_pOutput != NULL) {
		TSkkRuleTreeNodeOutput_vDestroy (pNode->_pOutput) ;
	}
	pNode->_pOutput	= pOutput ;
	return ;
}

struct CSkkRuleTreeNode*
TSkkRuleTreeNode_pGetBranchList (
	struct CSkkRuleTreeNode*		pRoot)
{
	return	pRoot != NULL? TSkkRuleTreeNode_pGetChild (pRoot) : NULL ;
}

struct CSkkRuleTreeNode*
TSkkRuleTreeNode_pSelectBranch (
	struct CSkkRuleTreeNode*	pRoot,
	int							nChar)
{
	struct CSkkRuleTreeNode*	pNode ;

	if (pRoot == NULL)
		return	NULL ;

	pNode	= TSkkRuleTreeNode_pGetChild (pRoot) ;
	while (pNode != NULL) {
		if (TSkkRuleTreeNode_iGetChar (pNode) == nChar)
			break ;
		pNode	= TSkkRuleTreeNode_pGetBrother (pNode) ;
	}
	return	pNode ;
}

/*	ԂĺATɐ(ǂȂȂ炻܂)
 *	̎_ł̃m[h ppResult ɓB
 */
int
TSkkRuleTreeNode_iSearchTree (
	struct CSkkRuleTreeNode*	pTree,
	LPCWSTR						pString,
	int							nString,
	struct CSkkRuleTreeNode**	ppResult)
{
	struct CSkkRuleTreeNode*	pParent ;
	struct CSkkRuleTreeNode*	pNode ;
	LPCWSTR		wptr ;
	int			nptr, nCH ;

	pNode		= pTree != NULL? TSkkRuleTreeNode_pGetChild (pTree) : NULL ;
	pParent		= pTree ;
	wptr		= pString ;
	nptr		= nString ;
	while (nptr > 0) {
		nCH		= *wptr ;
		/* Z푤̒TB*/
		while (pNode != NULL) {
			if (TSkkRuleTreeNode_iGetChar (pNode) == nCH)
				break ;
			pNode	= TSkkRuleTreeNode_pGetBrother (pNode) ;
		}
		if (pNode == NULL) {
			/* łBvĂƂ܂ŕԂB*/
			if (ppResult != NULL)
				*ppResult	= pParent ;
#if defined (DEBUG)
			_tprintf (TEXT ("iSeachTree: not found\n")) ;
#endif
			return	wptr - pString ;
		}
		wptr	++ ;
		nptr	-- ;
		if (nptr <= 0) {
			/* B*/
			if (ppResult != NULL)
				*ppResult	= pNode ;
#if defined (DEBUG)
			_tprintf (TEXT ("iSeachTree: found\n")) ;
			TSkkRuleTreeNode_vDebugOut (pNode) ;
			TSkkRuleTreeNode_vDebugOut (pParent) ;
#endif
			return	wptr - pString ;
		}
#if defined (DEBUG)
		_tprintf (TEXT ("iSearchTree: -> goto child\n")) ;
		TSkkRuleTreeNode_vDebugOut (pNode) ;
#endif
		pParent	= pNode ;
		pNode	= TSkkRuleTreeNode_pGetChild (pNode) ;
	}
	if (ppResult != NULL)
		*ppResult	= pTree ;
	return	0 ;
}

BOOL
TSkkRuleTreeNode_bAddRuleSP (
	struct CSkkRuleTreeNode**	ppTree,
	LPCWSTR						wstrPrefix,
	int							nPrefixLen,
	LPCWSTR						wstrNext,
	int							nNextLen,
	LPCWSTR						wstrHira,
	int							nHiraLen,
	LPCWSTR						wstrKata,
	int							nKataLen)
{
	struct CSkkRuleTreeNode*	pRoot ;
	struct CSkkRuleTreeNode*	pAddPoint ;
	struct CSkkRuleTreeNode*	pAddTree ;
	int							nOffset = 0 ;
	struct CSkkRuleTreeNodeOutput*	pOutput	= NULL ;

	/* prefix ݂Ȃ̂͋ȂBprefix Ȃ͔ԕ݂̂łB*/
	if (ppTree == NULL || wstrPrefix == NULL || nPrefixLen <= 0)
		return	FALSE ;

	pRoot		= *ppTree ;
	/* ԕpӂB original ̍\ɂ킹ĂB*/
	if (pRoot == NULL) {
		pRoot	= TSkkRuleTreeNode_pCreateInstance (0, NULL, 0, NULL, 0, NULL) ;
		if (pRoot == NULL)
			return	FALSE ;
		*ppTree	= pRoot ;
	}
	nOffset		= TSkkRuleTreeNode_iSearchTree (pRoot, wstrPrefix, nPrefixLen, &pAddPoint) ;
	if (nOffset < nPrefixLen) {
		/*	S̐VKo^łꍇB*/
		while (nOffset < (nPrefixLen - 1)) {
			/* r̈ʒu next  hira, kata ͎gȂB*/
			pAddTree	= TSkkRuleTreeNode_pCreateInstance (wstrPrefix [nOffset], wstrPrefix, nOffset + 1, NULL, 0, NULL) ;
			if (pAddTree == NULL)
				return	FALSE ;

			/*	addpoint  addtree ǉB\[gȂȂycB*/
			ASSERT (pAddPoint != NULL && pRoot != NULL) ;
			TSkkRuleTreeNode_vSetBrother (pAddTree, TSkkRuleTreeNode_pGetChild (pAddPoint)) ;
			TSkkRuleTreeNode_vSetChild (pAddPoint, pAddTree) ;
			pAddPoint	= pAddTree ;
			nOffset	++ ;
		}

		pOutput		= TSkkRuleTreeNodeOutput_pCreateStringPairInstance (wstrHira, nHiraLen, wstrKata, nKataLen) ;
		if (pOutput == NULL)
			return	FALSE ;
		pAddTree	= TSkkRuleTreeNode_pCreateInstance (wstrPrefix [nOffset], wstrPrefix, nPrefixLen, wstrNext, nNextLen, pOutput) ;
		if (pAddTree == NULL) {
			TSkkRuleTreeNodeOutput_vDestroy (pOutput) ;
			return	FALSE ;
		}

		/*	addpoint  addtree ǉB\[gȂȂycB*/
		ASSERT (pAddPoint != NULL && pRoot != NULL) ;
		TSkkRuleTreeNode_vSetBrother (pAddTree, TSkkRuleTreeNode_pGetChild (pAddPoint)) ;
		TSkkRuleTreeNode_vSetChild (pAddPoint, pAddTree) ;
	} else {
		pOutput		= TSkkRuleTreeNodeOutput_pCreateStringPairInstance (wstrHira, nHiraLen, wstrKata, nKataLen) ;
		if (pOutput == NULL)
			return	FALSE ;
		TSkkRuleTreeNode_vSetOutput (pAddPoint, pOutput) ;
	}
	return	TRUE ;
}

BOOL
TSkkRuleTreeNode_bAddRuleS (
	struct CSkkRuleTreeNode**	ppTree,
	LPCWSTR						wstrPrefix,
	int							nPrefixLen,
	LPCWSTR						wstrNext,
	int							nNextLen,
	LPCWSTR						wstrString,
	int							nString)
{
	struct CSkkRuleTreeNode*		pRoot ;
	struct CSkkRuleTreeNode*		pAddPoint ;
	struct CSkkRuleTreeNode*		pAddTree ;
	struct CSkkRuleTreeNodeOutput*	pOutput	= NULL ;
	int						nOffset = 0 ;

	/* prefix ݂Ȃ̂͋ȂBprefix Ȃ͔ԕ݂̂łB*/
	if (ppTree == NULL || wstrPrefix == NULL || nPrefixLen <= 0)
		return	FALSE ;

	pRoot		= *ppTree ;
	/* ԕpӂB original ̍\ɂ킹ĂB*/
	if (pRoot == NULL) {
		pRoot	= TSkkRuleTreeNode_pCreateInstance (0, NULL, 0, NULL, 0, NULL) ;
		if (pRoot == NULL)
			return	FALSE ;
		*ppTree	= pRoot ;
	}
	nOffset		= TSkkRuleTreeNode_iSearchTree (pRoot, wstrPrefix, nPrefixLen, &pAddPoint) ;
	if (nOffset < nPrefixLen) {
		/*	S̐VKo^łꍇB*/
		while (nOffset < (nPrefixLen - 1)) {
			/* r̈ʒu next  hira, kata ͎gȂB*/
			pAddTree	= TSkkRuleTreeNode_pCreateInstance (wstrPrefix [nOffset], wstrPrefix, nOffset + 1, NULL, 0, NULL) ;
			if (pAddTree == NULL)
				return	FALSE ;

			/*	addpoint  addtree ǉB\[gȂȂycB*/
			ASSERT (pAddPoint != NULL && pRoot != NULL) ;
			TSkkRuleTreeNode_vSetBrother (pAddTree, TSkkRuleTreeNode_pGetChild (pAddPoint)) ;
			TSkkRuleTreeNode_vSetChild (pAddPoint, pAddTree) ;
			pAddPoint	= pAddTree ;
			nOffset	++ ;
		}

		pOutput		= TSkkRuleTreeNodeOutput_pCreateStringInstance (wstrString, nString) ;
		if (pOutput == NULL)
			return	FALSE ;
		pAddTree	= TSkkRuleTreeNode_pCreateInstance (wstrPrefix [nOffset], wstrPrefix, nPrefixLen, wstrNext, nNextLen, pOutput) ;
		if (pAddTree == NULL) {
			TSkkRuleTreeNodeOutput_vDestroy (pOutput) ;
			return	FALSE ;
		}

		/*	addpoint  addtree ǉB\[gȂȂycB*/
		ASSERT (pAddPoint != NULL && pRoot != NULL) ;
		TSkkRuleTreeNode_vSetBrother (pAddTree, TSkkRuleTreeNode_pGetChild (pAddPoint)) ;
		TSkkRuleTreeNode_vSetChild (pAddPoint, pAddTree) ;
	} else {
		pOutput		= TSkkRuleTreeNodeOutput_pCreateStringInstance (wstrString, nString) ;
		if (pOutput == NULL)
			return	FALSE ;
		TSkkRuleTreeNode_vSetOutput (pAddPoint, pOutput) ;
	}
	return	TRUE ;
}

BOOL
TSkkRuleTreeNode_bAddRuleM (
	struct CSkkRuleTreeNode**	ppTree,
	LPCWSTR						wstrPrefix,
	int							nPrefixLen,
	LPCWSTR						wstrNext,
	int							nNextLen,
	int							nMacro) 
{
	struct CSkkRuleTreeNode*		pRoot ;
	struct CSkkRuleTreeNode*		pAddPoint ;
	struct CSkkRuleTreeNode*		pAddTree ;
	int						nOffset = 0 ;
	struct CSkkRuleTreeNodeOutput*	pOutput	= NULL ;

	/* prefix ݂Ȃ̂͋ȂBprefix Ȃ͔ԕ݂̂łB*/
	if (ppTree == NULL || wstrPrefix == NULL || nPrefixLen <= 0)
		return	FALSE ;

	pRoot		= *ppTree ;
	/* ԕpӂB original ̍\ɂ킹ĂB*/
	if (pRoot == NULL) {
		pRoot	= TSkkRuleTreeNode_pCreateInstance (0, NULL, 0, NULL, 0, NULL) ;
		if (pRoot == NULL)
			return	FALSE ;
		*ppTree	= pRoot ;
	}
	nOffset		= TSkkRuleTreeNode_iSearchTree (pRoot, wstrPrefix, nPrefixLen, &pAddPoint) ;
	if (nOffset < nPrefixLen) {
		/*	S̐VKo^łꍇB*/
		while (nOffset < (nPrefixLen - 1)) {
			/* r̈ʒu next  hira, kata ͎gȂB*/
			pAddTree	= TSkkRuleTreeNode_pCreateInstance (wstrPrefix [nOffset], wstrPrefix, nOffset + 1, NULL, 0, NULL) ;
			if (pAddTree == NULL)
				return	FALSE ;

			/*	addpoint  addtree ǉB\[gȂȂycB*/
			ASSERT (pAddPoint != NULL && pRoot != NULL) ;
			TSkkRuleTreeNode_vSetBrother (pAddTree, TSkkRuleTreeNode_pGetChild (pAddPoint)) ;
			TSkkRuleTreeNode_vSetChild (pAddPoint, pAddTree) ;
			pAddPoint	= pAddTree ;
			nOffset	++ ;
		}

		pOutput		= TSkkRuleTreeNodeOutput_pCreateMacroInstance (nMacro) ;
		if (pOutput == NULL)
			return	FALSE ;
		pAddTree	= TSkkRuleTreeNode_pCreateInstance (wstrPrefix [nOffset], wstrPrefix, nPrefixLen, wstrNext, nNextLen, pOutput) ;
		if (pAddTree == NULL) {
			TSkkRuleTreeNodeOutput_vDestroy (pOutput) ;
			return	FALSE ;
		}

		/*	addpoint  addtree ǉB\[gȂȂycB*/
		ASSERT (pAddPoint != NULL && pRoot != NULL) ;
		TSkkRuleTreeNode_vSetBrother (pAddTree, TSkkRuleTreeNode_pGetChild (pAddPoint)) ;
		TSkkRuleTreeNode_vSetChild (pAddPoint, pAddTree) ;
	} else {
		pOutput		= TSkkRuleTreeNodeOutput_pCreateMacroInstance (nMacro) ;
		if (pOutput == NULL)
			return	FALSE ;
		TSkkRuleTreeNode_vSetOutput (pAddPoint, pOutput) ;
	}
	return	TRUE ;
}

BOOL
TSkkRuleTreeNode_bDeleteRule (
	struct CSkkRuleTreeNode**	ppTree,
	LPCWSTR						pString,
	int							nString)
{
	struct CSkkRuleTreeNode*	pTree ;
	struct CSkkRuleTreeNode*	pParent ;
	struct CSkkRuleTreeNode*	pNode ;
	struct CSkkRuleTreeNode*	pPrev ;
	LPCWSTR		wptr ;
	int			nptr, nCH ;

	pTree	= *ppTree ;
	pNode	= pTree != NULL? TSkkRuleTreeNode_pGetChild (pTree) : NULL ;
	pParent	= pTree ;
	wptr	= pString ;
	nptr	= nString ;

	/*	܂ brother ւ̒TsB*/
	while (pNode != NULL && nptr > 0) {
		nCH		= *wptr ;
		pPrev	= NULL ;
		while (pNode != NULL) {
			if (TSkkRuleTreeNode_iGetChar (pNode) == nCH) 
				break ;
			pPrev	= pNode ;
			pNode	= TSkkRuleTreeNode_pGetBrother (pNode) ;
		}
		/* YȂB*/
		if (pNode == NULL)
			return	FALSE ;

		wptr	++ ;
		nptr	-- ;
		if (nptr <= 0) {
			/* ̈ʒuō폜B*/
			if (pPrev == NULL) {
				/* ẽN\ւB*/
				if (pParent != NULL) {
					TSkkRuleTreeNode_vSetChild (pParent, TSkkRuleTreeNode_pGetBrother (pNode)) ;
				} else {
					*ppTree				= TSkkRuleTreeNode_pGetBrother (pNode) ;
				}
				TSkkRuleTreeNode_vSetBrother (pNode, NULL) ;
			} else {
				/*	Z̃N\ւB
				 *	cqq͑SłB
				 */
				TSkkRuleTreeNode_vSetBrother (pPrev, TSkkRuleTreeNode_pGetBrother (pNode)) ;
				TSkkRuleTreeNode_vSetBrother (pNode, NULL) ;
			}
			TSkkRuleTreeNode_vDestroyTree (pNode) ;
			break ;
		}
		pParent	= pNode ;
		pNode	= TSkkRuleTreeNode_pGetChild (pNode) ;
	}
	return	TRUE ;
}

void
TSkkRuleTreeNode_vDestroyTree (
	struct CSkkRuleTreeNode*	pRoot)
{
	struct CSkkRuleTreeNode*	pBrother ;
	struct CSkkRuleTreeNode*	pChild ;

	if (pRoot == NULL)
		return ;

	pBrother	= TSkkRuleTreeNode_pGetBrother (pRoot) ;
	pChild		= TSkkRuleTreeNode_pGetChild (pRoot) ;
	TSkkRuleTreeNode_vSetBrother (pRoot, NULL) ;
	TSkkRuleTreeNode_vSetChild (pRoot, NULL) ;
	TSkkRuleTreeNode_vDestroy (pRoot) ;

	if (pBrother != NULL)
		TSkkRuleTreeNode_vDestroyTree (pBrother) ;
	if (pChild != NULL)
		TSkkRuleTreeNode_vDestroyTree (pChild) ;
	return ;
}

/*===============================================================
 *	public methods
 */


void
TSkkRuleTreeNode_vDebugOut (const struct CSkkRuleTreeNode* pNode)
{
	WCHAR	buf [256] ;
	WCHAR	bufPrefix [128], bufNextState [128] ;
	int		n ;

	if (pNode->_pPrefix != NULL && pNode->_nPrefix > 0) {
		n	= (pNode->_nPrefix < (ARRAYSIZE (bufPrefix) - 1))? pNode->_nPrefix : (ARRAYSIZE (bufPrefix) - 1) ;
		memcpy (bufPrefix, pNode->_pPrefix, n * sizeof (WCHAR)) ;
		bufPrefix [n]	= L'\0' ;
	} else {
		bufPrefix [0]	= L'\0' ;
	}
	if (pNode->_pNextState != NULL && pNode->_nNextState > 0) {
		n	= (pNode->_nNextState < (ARRAYSIZE (bufNextState) - 1))? pNode->_nNextState : (ARRAYSIZE (bufNextState) - 1) ;
		memcpy (bufNextState, pNode->_pNextState, n * sizeof (WCHAR)) ;
		bufNextState [n]	= L'\0' ;
	} else {
		bufNextState [0]	= L'\0' ;
	}
	if (pNode->_pOutput != NULL) {
		switch (TSkkRuleTreeNodeOutput_iGetType (pNode->_pOutput)) {
		case	NTYPE_STRINGPAIR:
			{
				LPCWSTR	pHira, pKata ;
				int		nHira, nKata ;
				WCHAR	bufHira [128], bufKata [128] ;

				if (! TSkkRuleTreeNodeOutput_bGetLString (pNode->_pOutput, &pHira, &nHira) ||
					! TSkkRuleTreeNodeOutput_bGetRString (pNode->_pOutput, &pKata, &nKata)) {
					break ;
				}
				if (pHira != NULL && nHira > 0) {
					n	= (nHira < (ARRAYSIZE (bufHira) - 1))? nHira : (ARRAYSIZE (bufHira) - 1) ;
					memcpy (bufHira, pHira, nHira * sizeof (WCHAR)) ;
					bufHira [n]	= L'\0' ;
				} else {
					bufHira [0]	= L'\0' ;
				}
				if (pKata != NULL && nKata > 0) {
					n	= (nKata < (ARRAYSIZE (bufKata) - 1))? nKata : (ARRAYSIZE (bufKata) - 1) ;
					memcpy (bufKata, pKata, nKata * sizeof (WCHAR)) ;
					bufKata [n]	= L'\0' ;
				} else {
					bufKata [0]	= L'\0' ;
				}
				wnsprintfW (buf, ARRAYSIZE (buf) - 1, L"Prefix(\"%s\"), Next(\"%s\"), Hira(\"%s\"), Kata(\"%s\")\n", bufPrefix, bufNextState, bufHira, bufKata) ;
				buf [ARRAYSIZE (buf) - 1]	= L'\0' ;
			}
			break ;

		case	NTYPE_STRING:
			{
				WCHAR	bufString [128] ;
				LPCWSTR	pString ;
				int		nString ;
				if (! TSkkRuleTreeNodeOutput_bGetString (pNode->_pOutput, &pString, &nString)) {
					break ;
				}
				if (pString != NULL && nString > 0) {
					n	= (nString < (ARRAYSIZE (bufString) - 1))? nString : (ARRAYSIZE (bufString) - 1) ;
					memcpy (bufString, pString, nString * sizeof (WCHAR)) ;
				} else {
					bufString [0]	= L'\0' ;
				}
				wnsprintfW (buf, ARRAYSIZE (buf) - 1, L"Prefix(\"%s\"), Next(\"%s\"), String(\"%s\")\n", bufPrefix, bufNextState, bufString) ;
				buf [ARRAYSIZE (buf) - 1]	= L'\0' ;
			}
			break ;

		case	NTYPE_MACRO:
			{
				int		nMacro ;
				if (! TSkkRuleTreeNodeOutput_bGetMacro (pNode->_pOutput, &nMacro)) {
					break ;
				}
				wnsprintfW (buf, ARRAYSIZE (buf) - 1, L"Prefix(\"%s\"), Next(\"%s\"), Macro(%d)\n", bufPrefix, bufNextState, nMacro) ;
				buf [ARRAYSIZE (buf) - 1]	= L'\0' ;
			}
			break ;
		default:
			break ;
		}
		wprintf (L"%s\n", buf) ;
	} else {
		wprintf (L"nil\n", buf) ;
	}
	return ;
}

struct TSkkBaseRuleT1 {
	LPCWSTR	_strState ;
	LPCWSTR	_strNext ;
	struct {
		LPCWSTR	_strKata ;
		LPCWSTR	_strHira ;
	}	_case ;
} ;

struct TSkkBaseRuleT2 {
	LPCWSTR	_strState ;
	LPCWSTR	_strNext ;
	LPCWSTR	_strOutput ;
} ;

struct TSkkBaseRuleT3 {
	LPCWSTR	_strState ;
	LPCWSTR	_strNext ;
	int		_nFunction ;
} ;

static struct TSkkBaseRuleT1	_rT1 []	= {
	{ L"a",		NULL,	{ L"A",	L"" } },
	{ L"bb",	L"b",	{ L"b",	L"" } },		{ L"ba",	NULL,	{ L"o",	L"" } },
	{ L"be",	NULL,	{ L"x",	L"" } },		{ L"bi",	NULL,	{ L"r",	L"" } },
	{ L"bo",	NULL,	{ L"{",	L"" } },		{ L"bu",	NULL,	{ L"u",	L"" } },
    { L"bya",	NULL,	{ L"r",	L"т" } },	{ L"bye",	NULL,	{ L"rF",	L"т" } },
	{ L"byi",	NULL,	{ L"rB",	L"т" } },	{ L"byo",	NULL,	{ L"r",	L"т" } },
	{ L"byu",	NULL,	{ L"r",	L"т" } },
    { L"cc",	L"c",	{ L"b",	L"" } },		{ L"cha",	NULL,	{ L"`",	L"" } },
    { L"che",	NULL,	{ L"`F",	L"" } },	{ L"chi",	NULL,	{ L"`",	L"" } },
    { L"cho",	NULL,	{ L"`",	L"" } },	{ L"chu",	NULL,	{ L"`",	L"" } },
	{ L"cya",	NULL,	{ L"`",	L"" } },	{ L"cye",	NULL,	{ L"`F",	L"" } },
	{ L"cyi",	NULL,	{ L"`B",	L"" } },	{ L"cyo",	NULL,	{ L"`",	L"" } },
	{ L"cyu",	NULL,	{ L"`",	L"" } },	{ L"dd",	L"d",	{ L"b",	L"" } },
	{ L"da",	NULL,	{ L"_",	L"" } },		{ L"de",	NULL,	{ L"f",	L"" } },
	{ L"dha",	NULL,	{ L"f",	L"ł" } },	{ L"dhe",	NULL,	{ L"fF",	L"ł" } },
	{ L"dhi",	NULL,	{ L"fB",	L"ł" } },	{ L"dho",	NULL,	{ L"f",	L"ł" } },
	{ L"dhu",	NULL,	{ L"f",	L"ł" } },	{ L"di",	NULL,	{ L"a",	L"" } },
	{ L"do",	NULL,	{ L"h",	L"" } },		{ L"du",	NULL,	{ L"d",	L"" } },
	{ L"dya",	NULL,	{ L"a",	L"" } },	{ L"dye",	NULL,	{ L"aF",	L"" } },
	{ L"dyi",	NULL,	{ L"aB",	L"" } },	{ L"dyo",	NULL,	{ L"a",	L"" } },
	{ L"dyu",	NULL,	{ L"a",	L"" } },
	{ L"e",		NULL,	{ L"G",	L"" } },
	{ L"ff",	L"f",	{ L"b",	L"" } },		{ L"fa",	NULL,	{ L"t@",	L"ӂ" } },
	{ L"fe",	NULL,	{ L"tF",	L"ӂ" } },	{ L"fi",	NULL,	{ L"tB",	L"ӂ" } },
	{ L"fo",	NULL,	{ L"tH",	L"ӂ" } },	{ L"fu",	NULL,	{ L"t",	L"" } },
	{ L"fya",	NULL,	{ L"t",	L"ӂ" } },	{ L"fye",	NULL,	{ L"tF",	L"ӂ" } },
	{ L"fyi",	NULL,	{ L"tB",	L"ӂ" } },	{ L"fyo",	NULL,	{ L"t",	L"ӂ" } },
	{ L"fyu",	NULL,	{ L"t",	L"ӂ" } },
	{ L"gg",	L"g",	{ L"b",	L"" } },		{ L"ga",	NULL,	{ L"K",	L"" } },
	{ L"ge",	NULL,	{ L"Q",	L"" } },		{ L"gi",	NULL,	{ L"M",	L"" } },
	{ L"go",	NULL,	{ L"S",	L"" } },		{ L"gu",	NULL,	{ L"O",	L"" } },
	{ L"gya",	NULL,	{ L"M",	L"" } },	{ L"gye",	NULL,	{ L"MF",	L"" } },
	{ L"gyi",	NULL,	{ L"MB",	L"" } },	{ L"gyo",	NULL,	{ L"M",	L"" } },
	{ L"gyu",	NULL,	{ L"M",	L"" } },
	//{ L"h", L"", { L"I", L"" } },
	{ L"ha",	NULL,	{ L"n",	L"" } },		{ L"he",	NULL,	{ L"w",	L"" } },
	{ L"hi",	NULL,	{ L"q",	L"" } },		{ L"ho",	NULL,	{ L"z",	L"" } },
	{ L"hu",	NULL,	{ L"t",	L"" } },		{ L"hya",	NULL,	{ L"q",	L"Ђ" } },
	{ L"hye",	NULL,	{ L"qF",	L"Ђ" } },	{ L"hyi",	NULL,	{ L"qB",	L"Ђ" } },
	{ L"hyo",	NULL,	{ L"q",	L"Ђ" } },	{ L"hyu",	NULL,	{ L"q",	L"Ђ" } },
	{ L"i",		NULL,	{ L"C",	L"" } },		{ L"jj",	L"j",	{ L"b",	L"" } },
	{ L"ja",	NULL,	{ L"W",	L"" } },	{ L"je",	NULL,	{ L"WF",	L"" } },
	{ L"ji",	NULL,	{ L"W",	L"" } },		{ L"jo",	NULL,	{ L"W",	L"" } },
	{ L"ju",	NULL,	{ L"W",	L"" } },	{ L"jya",	NULL,	{ L"W",	L"" } },
	{ L"jye",	NULL,	{ L"WF",	L"" } },	{ L"jyi",	NULL,	{ L"WB",	L"" } },
	{ L"jyo",	NULL,	{ L"W",	L"" } },	{ L"jyu",	NULL,	{ L"W",	L"" } },
	{ L"kk",	L"k",	{ L"b",	L"" } },		{ L"ka",	NULL,	{ L"J",	L"" } },
	{ L"ke",	NULL,	{ L"P",	L"" } },		{ L"ki",	NULL,	{ L"L",	L"" } },
	{ L"ko",	NULL,	{ L"R",	L"" } },		{ L"ku",	NULL,	{ L"N",	L"" } },
	{ L"kya",	NULL,	{ L"L",	L"" } },	{ L"kye",	NULL,	{ L"LF",	L"" } },
	{ L"kyi",	NULL,	{ L"LB",	L"" } },	{ L"kyo",	NULL,	{ L"L",	L"" } },
	{ L"kyu",	NULL,	{ L"L",	L"" } },
	{ L"ma",	NULL,	{ L"}",	L"" } },		{ L"me",	NULL,	{ L"",	L"" } },
	{ L"mi",	NULL,	{ L"~",	L"" } },		{ L"mo",	NULL,	{ L"",	L"" } },
	{ L"mu",	NULL,	{ L"",	L"" } },		{ L"mya",	NULL,	{ L"~",	L"݂" } },
	{ L"mye",	NULL,	{ L"~F",	L"݂" } },	{ L"myi",	NULL,	{ L"~B",	L"݂" } },
	{ L"myo",	NULL,	{ L"~",	L"݂" } },	{ L"myu",	NULL,	{ L"~",	L"݂" } },
	{ L"n",		NULL,	{ L"",	L"" } },		{ L"n'",	NULL,	{ L"",	L"" } },
	{ L"na",	NULL,	{ L"i",	L"" } },		{ L"ne",	NULL,	{ L"l",	L"" } },
	{ L"ni",	NULL,	{ L"j",	L"" } },		{ L"nn",	NULL,	{ L"",	L"" } },
	{ L"no",	NULL,	{ L"m",	L"" } },		{ L"nu",	NULL,	{ L"k",	L"" } },
	{ L"nya",	NULL,	{ L"j",	L"ɂ" } },	{ L"nye",	NULL,	{ L"jF",	L"ɂ" } },
	{ L"nyi",	NULL,	{ L"jB",	L"ɂ" } },	{ L"nyo",	NULL,	{ L"j",	L"ɂ" } },
	{ L"nyu",	NULL,	{ L"j",	L"ɂ" } },
	{ L"o",		NULL,	{ L"I",	L"" } },
	{ L"pp",	L"p",	{ L"b",	L"" } },		{ L"pa",	NULL,	{ L"p",	L"" } },
	{ L"pe",	NULL,	{ L"y",	L"" } },		{ L"pi",	NULL,	{ L"s",	L"" } },
	{ L"po",	NULL,	{ L"|",	L"" } },		{ L"pu",	NULL,	{ L"v",	L"" } },
	{ L"pya",	NULL,	{ L"s",	L"҂" } },	{ L"pye",	NULL,	{ L"sF",	L"҂" } },
	{ L"pyi",	NULL,	{ L"sB",	L"҂" } },	{ L"pyo",	NULL,	{ L"s",	L"҂" } },
	{ L"pyu",	NULL,	{ L"s",	L"҂" } },
	{ L"rr",	L"r",	{ L"b",	L"" } },		{ L"ra",	NULL,	{ L"",	L"" } },
	{ L"re",	NULL,	{ L"",	L"" } },		{ L"ri",	NULL,	{ L"",	L"" } },
	{ L"ro",	NULL,	{ L"",	L"" } },		{ L"ru",	NULL,	{ L"",	L"" } },
	{ L"rya",	NULL,	{ L"",	L"" } },	{ L"rye",	NULL,	{ L"F",	L"肥" } },
	{ L"ryi",	NULL,	{ L"B",	L"股" } },	{ L"ryo",	NULL,	{ L"",	L"" } },
	{ L"ryu",	NULL,	{ L"",	L"" } },
	{ L"ss",	L"s",	{ L"b",	L"" } },		{ L"sa",	NULL,	{ L"T",	L"" } },
	{ L"se",	NULL,	{ L"Z",	L"" } },		{ L"sha",	NULL,	{ L"V",	L"" } },
	{ L"she",	NULL,	{ L"VF",	L"" } },	{ L"shi",	NULL,	{ L"V",	L"" } },
	{ L"sho",	NULL,	{ L"V",	L"" } },	{ L"shu",	NULL,	{ L"V",	L"" } },
	{ L"si",	NULL,	{ L"V",	L"" } },		{ L"so",	NULL,	{ L"\",	L"" } },
	{ L"su",	NULL,	{ L"X",	L"" } },		{ L"sya",	NULL,	{ L"V",	L"" } },
	{ L"sye",	NULL,	{ L"VF",	L"" } },	{ L"syi",	NULL,	{ L"VB",	L"" } },
	{ L"syo",	NULL,	{ L"V",	L"" } },	{ L"syu",	NULL,	{ L"V",	L"" } },
	{ L"tt",	L"t",	{ L"b",	L"" } },		{ L"ta",	NULL,	{ L"^",	L"" } },
	{ L"te",	NULL,	{ L"e",	L"" } },		{ L"tha",	NULL,	{ L"e@",	L"Ă" } },
	{ L"the",	NULL,	{ L"eF",	L"Ă" } },	{ L"thi",	NULL,	{ L"eB",	L"Ă" } },
	{ L"tho",	NULL,	{ L"e",	L"Ă" } },	{ L"thu",	NULL,	{ L"e",	L"Ă" } },
	{ L"ti",	NULL,	{ L"`",	L"" } },		{ L"to",	NULL,	{ L"g",	L"" } },
	{ L"tsu",	NULL,	{ L"c",	L"" } },		{ L"tu",	NULL,	{ L"c",	L"" } },
	{ L"tya",	NULL,	{ L"`",	L"" } },	{ L"tye",	NULL,	{ L"`F",	L"" } },
	{ L"tyi",	NULL,	{ L"`B",	L"" } },	{ L"tyo",	NULL,	{ L"`",	L"" } },
	{ L"tyu",	NULL,	{ L"`",	L"" } },
	{ L"u",		NULL,	{ L"E",	L"" } },
	{ L"vv",	L"v",	{ L"b",	L"" } },		{ L"va",	NULL,	{ L"@",	L"J" } },
	{ L"ve",	NULL,	{ L"F",	L"J" } },	{ L"vi",	NULL,	{ L"B",	L"J" } },
	{ L"vo",	NULL,	{ L"H",	L"J" } },	{ L"vu",	NULL,	{ L"",	L"J" } },
	{ L"ww",	L"w",	{ L"b",	L"" } },		{ L"wa",	NULL,	{ L"",	L"" } },
	{ L"we",	NULL,	{ L"EF",	L"" } },	{ L"wi",	NULL,	{ L"EB",	L"" } },
	{ L"wo",	NULL,	{ L"",	L"" } },		{ L"wu",	NULL,	{ L"E",	L"" } },
	{ L"xx",	L"x",	{ L"b",	L"" } },		{ L"xa",	NULL,	{ L"@",	L"" } },
	{ L"xe",	NULL,	{ L"F",	L"" } },		{ L"xi",	NULL,	{ L"B",	L"" } },
	{ L"xka",	NULL,	{ L"",	L"" } },		{ L"xke",	NULL,	{ L"",	L"" } },
	{ L"xo",	NULL,	{ L"H",	L"" } },		{ L"xtsu",	NULL,	{ L"b",	L"" } },
	{ L"xtu",	NULL,	{ L"b",	L"" } },		{ L"xu",	NULL,	{ L"D",	L"" } },
	{ L"xwa",	NULL,	{ L"",	L"" } },		{ L"xwe",	NULL,	{ L"",	L"" } },
	{ L"xwi",	NULL,	{ L"",	L"" } },		{ L"xya",	NULL,	{ L"",	L"" } },
	{ L"xyo",	NULL,	{ L"",	L"" } },		{ L"xyu",	NULL,	{ L"",	L"" } },
	{ L"yy",	L"y",	{ L"b",	L"" } },		{ L"ya",	NULL,	{ L"",	L"" } },
	{ L"ye",	NULL,	{ L"CF",	L"" } },	{ L"yo",	NULL,	{ L"",	L"" } },
	{ L"yu",	NULL,	{ L"",	L"" } },
	{ L"zz",	L"z",	{ L"b",	L"" } },		{ L"za",	NULL,	{ L"U",	L"" } },
	{ L"ze",	NULL,	{ L"[",	L"" } },		{ L"zi",	NULL,	{ L"W",	L"" } },
	{ L"zo",	NULL,	{ L"]",	L"" } },		{ L"zu",	NULL,	{ L"Y",	L"" } },
	{ L"zya",	NULL,	{ L"W",	L"" } },	{ L"zye",	NULL,	{ L"WF",	L"" } },
	{ L"zyi",	NULL,	{ L"WB",	L"" } },	{ L"zyo",	NULL,	{ L"W",	L"" } },
	{ L"zyu",	NULL,	{ L"W",	L"" } },
} ;

static struct TSkkBaseRuleT2	_rT2 []	= {
	{ L"z,",	NULL,	L"d" },	{ L"z-",	NULL,	L"`" },    { L"z.",	NULL,	L"c" },
	{ L"z/",	NULL,	L"E" },    { L"z[",	NULL,	L"w" },    { L"z]",	NULL,	L"x" },
    { L"zh",	NULL,	L"" },    { L"zj",	NULL,	L"" },    { L"zk",	NULL,	L"" },
	{ L"zl",	NULL,	L"" },    { L"-",		NULL,	L"[" },    { L":",		NULL,	L"F" },
    { L";",		NULL,	L"G" },    { L"?",		NULL,	L"H" },    { L"[",		NULL,	L"u" },
	{ L"]",		NULL,	L"v" },
} ;

static struct TSkkBaseRuleT3	_rT3 []	= {
	{ L".",		NULL,	NFUNC_SKK_CURRENT_KUTEN },			{ L",",		NULL,	NFUNC_SKK_CURRENT_TOUTEN },
	{ L"l",		NULL,	NFUNC_SKK_LATIN_MODE },				{ L"q",		NULL,	NFUNC_SKK_TOGGLE_KANA },
	{ L"L",		NULL,	NFUNC_SKK_JISX0208_LATIN_MODE },	{ L"Q",		NULL,	NFUNC_SKK_SET_HENKAN_POINT_SUBR },
	{ L"X",		NULL,	NFUNC_SKK_PURGE_FROM_JISYO },		{ L"/",		NULL,	NFUNC_SKK_ABBREV_MODE },
	{ L"$",		NULL,	NFUNC_SKK_DISPLAY_CODE_FOR_CHAR_AT_POINT },
	{ L"@",		NULL,	NFUNC_SKK_TODAY },					{ L"\\",	NULL,	NFUNC_SKK_INPUT_BY_CODE_OR_MENU },
	{ L"\r",	NULL,	NFUNC_SKK_KAKUTEI },
} ;

BOOL
TSkkRuleTree_bInitializeDefault (
	struct CSkkRuleTreeNode**			ppTree)
{
	struct CSkkRuleTreeNode*	pRoot ;
	int		i, nPrefix, nNextState ;

	if (ppTree == NULL)
		return	FALSE ;

	pRoot	= NULL ;
	for (i = 0 ; i < ARRAYSIZE (_rT1) ; i ++) {
		int	nHira, nKata ;

		nPrefix		= (_rT1 [i]._strState      != NULL)? lstrlenW (_rT1 [i]._strState)      : 0 ;
		nNextState	= (_rT1 [i]._strNext       != NULL)? lstrlenW (_rT1 [i]._strNext)       : 0 ;
		nHira		= (_rT1 [i]._case._strHira != NULL)? lstrlenW (_rT1 [i]._case._strHira) : 0 ;
		nKata		= (_rT1 [i]._case._strKata != NULL)? lstrlenW (_rT1 [i]._case._strKata) : 0 ;
		if (! TSkkRuleTreeNode_bAddRuleSP (&pRoot, _rT1 [i]._strState, nPrefix, _rT1 [i]._strNext, nNextState, _rT1 [i]._case._strHira, nHira, _rT1 [i]._case._strKata, nKata)) {
			TSkkRuleTreeNode_vDestroyTree (pRoot) ;
			return	FALSE ;
		}
	}
	for (i = 0 ; i < ARRAYSIZE (_rT2) ; i ++) {
		int	nOutput ;
		nPrefix		= (_rT2 [i]._strState  != NULL)? lstrlenW (_rT2 [i]._strState)  : 0 ;
		nNextState	= (_rT2 [i]._strNext   != NULL)? lstrlenW (_rT2 [i]._strNext)   : 0 ;
		nOutput		= (_rT2 [i]._strOutput != NULL)? lstrlenW (_rT2 [i]._strOutput) : 0 ;
		if (! TSkkRuleTreeNode_bAddRuleS (&pRoot, _rT2 [i]._strState, nPrefix, _rT2 [i]._strNext, nNextState, _rT2 [i]._strOutput, nOutput)) {
			TSkkRuleTreeNode_vDestroyTree (pRoot) ;
			return	FALSE ;
		}
	}
	for (i = 0 ; i < ARRAYSIZE (_rT3) ; i ++) {
		nPrefix		= (_rT3 [i]._strState  != NULL)? lstrlenW (_rT3 [i]._strState)  : 0 ;
		nNextState	= (_rT3 [i]._strNext   != NULL)? lstrlenW (_rT3 [i]._strNext)   : 0 ;
		if (! TSkkRuleTreeNode_bAddRuleM (&pRoot, _rT3 [i]._strState, nPrefix, _rT3 [i]._strNext, nNextState, _rT3 [i]._nFunction)) {
			TSkkRuleTreeNode_vDestroyTree (pRoot) ;
			return	FALSE ;
		}
	}
	*ppTree	= pRoot ;
	return	TRUE ;
}

#if defined (UNIT_TEST)
/*	P̃eXĝ߂̃R[hB
 */
int
_tmain (void)
{
	struct CSkkRuleTreeNode*	pRoot	= NULL ;
	struct CSkkRuleTreeNode*	pNode	= NULL ;
	int		i, n ;
	char	buf [12] ;

	/*
struct TSkkBaseRuleT1 {
	LPCTSTR	_strState ;
	LPCTSTR	_strNext ;
	struct {
		LPCTSTR	_strHira ;
		LPCTSTR	_strKata ;
	}	_case ;
} ;*/
	_tsetlocale (LC_CTYPE, TEXT ("")) ;

	/* debugger  process  attach 邽߂̗]TB*/
	_tprintf (TEXT ("> ")) ;
	fflush (stdout) ;
	fgets (buf, 10, stdin) ;

	for (i = 0 ; i < ARRAYSIZE (_rT1) ; i ++) {
		int	nPrefix, nNextState, nHira, nKata ;

		nPrefix		= (_rT1 [i]._strState      != NULL)? lstrlenW (_rT1 [i]._strState)      : 0 ;
		nNextState	= (_rT1 [i]._strNext       != NULL)? lstrlenW (_rT1 [i]._strNext)       : 0 ;
		nHira		= (_rT1 [i]._case._strHira != NULL)? lstrlenW (_rT1 [i]._case._strHira) : 0 ;
		nKata		= (_rT1 [i]._case._strKata != NULL)? lstrlenW (_rT1 [i]._case._strKata) : 0 ;
		if (! TSkkRuleTreeNode_bAddRule (&pRoot, _rT1 [i]._strState, nPrefix, _rT1 [i]._strNext, nNextState, _rT1 [i]._case._strHira, nHira, _rT1 [i]._case._strKata, nKata)) {
			_tprintf (TEXT ("TSkkRuleTreeNode_bAddRule: failed. (%d)\n"), i) ;
			break ;
		}
	}
	if (pRoot != NULL) {
		_tprintf (TEXT ("Root ...\n")) ;
		TSkkRuleTreeNode_vDebugOut (pRoot) ;
	}
	n	= TSkkRuleTreeNode_iSearchTree (pRoot, L"b", 1, &pNode) ;
	_tprintf (TEXT ("Search ... \"b\" ... hit(%d)\n"), n) ;
	if (pNode != NULL) {
		TSkkRuleTreeNode_vDebugOut (pNode) ;
	}
	n	= TSkkRuleTreeNode_iSearchTree (pRoot, L"fya", 3, &pNode) ;
	_tprintf (TEXT ("Search ... \"fya\" ... hit(%d)\n"), n) ;
	if (pNode != NULL) {
		TSkkRuleTreeNode_vDebugOut (pNode) ;
	}
	return	0 ;
}
#endif



